from math import sin, cos, pi
from gaft import GAEngine
from gaft.components import BinaryIndividual
from gaft.components import Population
from gaft.operators import RouletteWheelSelection
from gaft.operators import TournamentSelection
from gaft.operators import UniformCrossover
from gaft.operators import FlipBitBigMutation

# Built-in best fitness analysis.
from gaft.analysis.fitness_store import FitnessStore
from gaft.analysis.console_output import ConsoleOutput

def gene_algo_corr(dict_tGCN,genetic_code_number,RSCU_df,bacteria, size_pop,generation_number):
  
    # Define population.
    indv_template = BinaryIndividual(ranges=[(0, 1),(0,1),(0,1),(0,1),(0,1)], eps=0.001)
    population = Population(indv_template=indv_template, size=size_pop).init()

    # Create genetic operators.
    selection = RouletteWheelSelection()
    crossover = UniformCrossover(pc=0.5, pe=0.25)
    mutation = FlipBitBigMutation(pm=0.1, pbm=0.5, alpha=0.51)

    # Create genetic algorithm engine.
    # Here we pass all built-in analysis to engine constructor.
    engine = GAEngine(population=population, selection=selection,
                    crossover=crossover, mutation=mutation,
                    analysis=[ConsoleOutput, FitnessStore])


    #########################################
    # Define fitness function.
    @engine.fitness_register
    #@engine.minimize
    def fitness(indv):
        from gtAI import new_flow 

        xug, xci, xai, xgu, xal = indv.solution

        Sug = xug
        Sci = xci
        Sai = xai
        Sgu = xgu
        Sal = xal
        
        Wi_df = new_flow.wi_tai_calc(dict_tGCN ,Sug=xug, Sci=xci, Sai=xai, Sgu=xgu, Sal=xal,genetic_code_number=genetic_code_number ,bacteria=bacteria)
        
        RSCU_wai_corr = new_flow.corr_result(Wi_df,RSCU_df)
        
        corr = float(RSCU_wai_corr["Wi"][0]) 
        if corr:
            return corr
        elif not corr:
            return 0
    
       

    engine.run(ng=generation_number)







import itertools 
import pandas as pd
import urllib
import pandas as pd
from pandas import DataFrame
import numpy as np


def tRNADB_CE(url):
    """
    Get the tRNA genes count from tRNADB-CE database

    Args:

        url (string): a url to anticodon table for organism from tRNADB_CE database (http://trna.ie.niigata-u.ac.jp/cgi-bin/trnadb/index.cgi)

    Returns:

        A dictionary of each anticodon and its gene count
    
    Raises:

        ValueError if the URL, not a valid for tRNADB-CE database
    
    Example:

        > tRNADB_CE("trna.ie.niigata-u.ac.jp/cgi-bin/trnadb/whole_anticodon.cgi?GID=|CP001631&DTYPE=CMP&VTYPE=1")
        # Return an anticodon table of Acidimicrobium ferrooxidans DSM 10331

    """

    if "http://trna.ie.niigata-u.ac.jp" not in url:
        raise ValueError ("url is not includes in tRNADB-CE database")
    
    all_trna_df = DataFrame()

    read = pd.read_html(url)
    read1 = read[1].iloc[3:,3:5]
    read1.columns = ["anti_codon", "count"]

    read2 = read[1].iloc[3:,6:8]
    read2.columns = ["anti_codon", "count"]

    read3 = read[1].iloc[3:,9:11]
    read3.columns = ["anti_codon", "count"]

    read4 = read[1].iloc[3:,12:14]
    read4.columns = ["anti_codon", "count"]


    all_trna_df = all_trna_df.append(read1,ignore_index=True, sort=False)
    all_trna_df = all_trna_df.append(read2,ignore_index=True, sort=False)
    all_trna_df = all_trna_df.append(read3,ignore_index=True, sort=False)
    all_trna_df = all_trna_df.append(read4,ignore_index=True, sort=False)


    A_to_I_anticodon = ["AGA","AAG","AGG","ACG","AAU","AGU","AAC","AGC","ACC"]  #codons which will convert its A to I in Wobble pos.
    anti_codon_dict = {}
    for codon, count in zip( list(all_trna_df["anti_codon"]), list(all_trna_df["count"]) ):
        codon = codon.replace("T","U")
        if codon in A_to_I_anticodon:
            codon = list(codon)
            codon[0] = "I"
            codon = "".join(codon)
        else:
            pass
        anti_codon_dict[codon] = int(count)

    #remove anti_codons with A in the first postion # caz it not recognize ani codon
    for i_anticodon in list(anti_codon_dict.keys()):
        if i_anticodon[0] == "A":
            del anti_codon_dict[i_anticodon]

    if len(anti_codon_dict) == 0:
        raise ValueError ("No tables were found in this URL")
    else:
        return anti_codon_dict

########################################################
              #########################
########################################################

def GtRNAdb(url,start_from_table):
    """
        Get the tRNA genes count from GtRNAdb database

    Args:

        url (string): a url to anticodon table for organism from GtRNAdb database (http://gtrnadb.ucsc.edu/)
        start_from_table (int): a number indicates where the anticodon table starts on the GtRNAdb page

    Returns:

        A dictionary of each anticodon and its gene count
    
    Raises:

        ValueError if the URL, not a valid for GtRNAdb database

    Example:

        #example 1

        > GtRNAdb("http://gtrnadb.ucsc.edu/genomes/eukaryota/Hsapi19/",start_from_table = 3)
        # Return an anticodon table of Homo sapiens (GRCh37/hg19)
        # Since there are 7 tables ( see: http://gtrnadb.ucsc.edu/genomes/eukaryota/Hsapi19/ ) 
        # So start_from_table parameter should be the number where the table of the actual count of anticodons starts.
        #Here its table number 3 (start from 0) ( so, start_from_table = 3 )

        #example 2

        > GtRNAdb("http://gtrnadb.ucsc.edu/genomes/eukaryota/Scere3/",start_from_table = 2)
        # Return an anticodon table of Saccharomyces cerevisiae S288c
        # Since there are 6 tables ( see: http://gtrnadb.ucsc.edu/genomes/eukaryota/Scere3/ ) 
        # So start_from_table parameter should be the number where the table of the actual count of anticodons starts.
        #Here its table number 2 (start from 0) ( so, start_from_table = 2 )

        #example 3

        > GtRNAdb("http://gtrnadb.ucsc.edu/genomes/bacteria/Esch_coli_K_12_MG1655/",start_from_table = 1)
        # Return an anticodon table of Escherichia coli str. K-12 substr. MG1655
        # Since there are 5 tables ( see: http://gtrnadb.ucsc.edu/genomes/bacteria/Esch_coli_K_12_MG1655/ ) 
        # So start_from_table parameter should be the number where the table of the actual count of anticodons starts.
        #Here its table number 1 (start from 0) ( so, start_from_table = 1 )

    """

    if start_from_table == 1 or start_from_table == 2 or start_from_table == 3:
        pass
    else:
        raise TypeError ("start_from_table should by 1 or 2 or 3 ")


    if "gtrnadb.ucsc.edu" not in url:
        raise ValueError ("url is not includes in GtRNAdb database")

    tables = pd.read_html(url)
    anti_codon_dict = {}

    def SPLIT(x):
        return x.split(" ")

    A_to_I_anticodon = ["AGA","AAG","AGG","ACG","AAU","AGU","AAC","AGC","ACC"]  #codons which will convert its A to I in Wobble pos.
    cat_anti_codon = ()
    for i in range(start_from_table,len(tables)):
        tables_new = tables[i]
    
        for j in range(0, len(tables_new.index)):
            overview = tables_new.iloc[j,1:] 
            codon_anticodon = list ( map(str, overview) )    
            codon_anticodon = list ( map(SPLIT, codon_anticodon) ) 
              
            #anti_codon with A that will convert to I
            for i_codon in codon_anticodon:
                if not i_codon[0].isdigit() and i_codon[0] != "&nbsp":
                    i_codon_rna = i_codon[0].replace("T","U")
                    if i_codon_rna in A_to_I_anticodon:
                        i_codon_rna = list(i_codon_rna)
                        i_codon_rna[0] = "I"
                        i_codon_rna = "".join(i_codon_rna)

                    try:
                        if i_codon_rna == "CAU" and "/" in i_codon[1]:
                            cat_anti_codon = cat_anti_codon + ( i_codon[1].split("/")[0] , )
                            cat_anti_codon = cat_anti_codon + ( i_codon[1].split("/")[1] , )
                        if i_codon_rna == "CAU" and "/" not in i_codon[1]:
                            cat_anti_codon = cat_anti_codon + ( i_codon[1] ,)
                        
                        
                        
                        anti_codon_dict[i_codon_rna] = int(i_codon[1])

                    except:
                        anti_codon_dict[i_codon_rna] = 0

    def seeifnumber(x):
        # take list and return a tuble with only number element
        result = ()
        for i in x:
            try:
                int(i)
                result = result + (int(i),)
            except:
                pass
        return (result)


    try:
        anti_codon_dict["CAU"] = sum(seeifnumber(cat_anti_codon))
    except:
        pass

    dict_anti_codons = {}
    
    #remove anti_codons with A in the first postion # caz it not recognize ani codon
    for i_anticodon in list(anti_codon_dict.keys()):
        if i_anticodon[0] == "A":
            del anti_codon_dict[i_anticodon]

    if len(anti_codon_dict) == 0:
        raise ValueError ("No tables were found in this URL")
    else:
        return anti_codon_dict

########################################################
              #########################
########################################################

def dict_codon_anticodon(anti_codon_dict):
    """
    Identify all potential anticodons for each codon 

    Args:

        anti_codon_dict (dict): a dictionary of all anticodons for an organism ( returned from ( tRNADB_CE() ) or ( GtRNAdb() ) )

    Returns:

        A dictionary of all potential anticodons for each codon 
    
    Raises:

        ValueError if the length of anti_codon_dict equal to zero
    
    Example:

        > anticodon_dict = tRNADB_CE("trna.ie.niigata-u.ac.jp/cgi-bin/trnadb/whole_anticodon.cgi?GID=|CP001631&DTYPE=CMP&VTYPE=1")
        # Return an anticodon table of Acidimicrobium ferrooxidans DSM 10331
        > dict_codon_anticodon ( anticanticodon_dictodon )

    """
    if len(anti_codon_dict) == 0:
        raise ValueError ("The dictionary can not be with length equal to zero")

    antiWC = { "A":"U", "U":"A", "C":"G", "G":"C" }
    antiWP = { "U":"G", "C":"I", "A":"I", "G":"U" }

    codon = ( "".join(i) for i in itertools.product('AUGC', repeat=3) )

    dic_codon_anticodon ={}

    #convert the codon to anticodon if anticodon in anti_codon_dict
    for i in codon:
        anti_codon_tuble = ( )

        if i[2] == "U":
            wc = antiWC[i[0]] + antiWC[i[1]] + "I"
            wc = wc[::-1]
        else:
            wc = antiWC[i[0]] + antiWC[i[1]] + antiWC[i[2]]
            wc = wc[::-1]
            
        if wc in anti_codon_dict:
            anti_codon_tuble = anti_codon_tuble + (wc, )
        else:
            pass
        #######

        wp = antiWC[i[0]] + antiWC[i[1]] + antiWP[i[2]] 
        wp = wp[::-1]

        if wp in anti_codon_dict:
            anti_codon_tuble = anti_codon_tuble + (wp, )
        else:
            pass

        if len(anti_codon_tuble) != 0: 
            dic_codon_anticodon[i] = ( anti_codon_tuble )
        else:
            pass

    return dic_codon_anticodon

########################################################
              #########################
########################################################

def dict_codon_anticodon_count(dic_codon_anticodon,dict_tGCN_main, bacteria = False):

    """ 
    Merge anticodon-codon dictionary with each anticodon tRNA gene copy number.

    Args:

        dic_codon_anticodon (dict): A dictionary of all potential anticodons for each codon returned from dict_codon_anticodon() function.

        dict_tGCN_main (dict): a dictionary of all anticodons for an organism ( returned from ( tRNADB_CE() function ) or ( GtRNAdb() function ) )

        bacteria (bool): True If the tested organism is prokaryotic or archaea, else equal to False ( default = False )

    Returns:
        
        a merged dictionary of anticodon-codon with each anticodon tRNA gene copy number.
    
    Raises:

        ValueError if the length of anti_codon_dict equal to zero
        ValueError if the length of dict_tGCN_main equal to zero

    Example:

        > anticodon_dict = tRNADB_CE("trna.ie.niigata-u.ac.jp/cgi-bin/trnadb/whole_anticodon.cgi?GID=|CP001631&DTYPE=CMP&VTYPE=1")
        # Return an anticodon table of Acidimicrobium ferrooxidans DSM 10331
        > anticodon_codon = dict_codon_anticodon ( anticanticodon_dictodon )
        > dict_codon_anticodon_count(anticodon_codon,anticodon_dict,bacteria = True)
    """

    if len(dic_codon_anticodon) == 0 or len(dict_tGCN_main) == 0:
        raise ValueError ("Both dic_codon_anticodon or dict_tGCN_main can not be with length equal to zero")

    dict_anticodon_number = {}
    for i in dic_codon_anticodon:
        temp = {}
        for j in range(len(dic_codon_anticodon[i])):
            
            
            temp[ dic_codon_anticodon[i][j] ]  = dict_tGCN_main[dic_codon_anticodon[i][j]] 

        dict_anticodon_number[i] = temp

    if bacteria:
        try:
            dict_anticodon_number["AUA"].update( { "CAU":dict_tGCN_main["CAU"] } )
        except:
            pass

    return dict_anticodon_number

########################################################
              #########################
########################################################
   
def abs_Wi(dict_anticodon_number, Sug,Sci,Sai,Sgu,Sal, bacteria=False):
    """
    Calculate the absolute adaptiveness values for each codon.

    Args:

        dict_anticodon_number (dict): a merged dictionary of anticodon-codon with each anticodon tRNA gene copy number returned from dict_codon_anticodon_count() function.

        Sug (int): the S-value for codon with (U) in the third position and (G) in first anticodon position

        Sci (int): the S-value for codon with (C) in the third position and (I) in first anticodon position

        Sai (int): the S-value for codon with (A) in the third position and (I) in first anticodon position

        Sgu (int): the S-value for codon with (G) in the third position and (U) in first anticodon position

        Sal (int): the S-value for codon with (A) in the third position and (L) in first anticodon position ( if bacteria = True)

        bacteria (bool): True If the tested organism is prokaryotic or archaea, else equal to False ( default = False )

    Returns:
        
        A dictionary of the absolute adaptiveness values for each codon

    Note:

        All Sij values (as Sug) should be a number from 0 to 1

    Raises:

        ValueError if the length of dict_anticodon_number equal to zero
        ValueError if any of Sij values are less than 0 or greater than 1

    Example:

        > anticodon_dict = tRNADB_CE("trna.ie.niigata-u.ac.jp/cgi-bin/trnadb/whole_anticodon.cgi?GID=|CP001631&DTYPE=CMP&VTYPE=1")
        # Return an anticodon table of Acidimicrobium ferrooxidans DSM 10331
        > anticodon_codon = dict_codon_anticodon ( anticanticodon_dictodon )
        > dict_codon_anticodon_count = dict_codon_anticodon_count(anticodon_codon,anticodon_dict,bacteria = True)
        > abs_Wi(dict_codon_anticodon_count, Sug=1, Sci=1, Sai=1, Sgu=1, Sal=1, bacteria=True) 
    """

    if len(dict_anticodon_number) == 0:
        raise ValueError ("dict_anticodon_number can not be with length equal to zero")

    if  0 > Sug > 1 or 0 > Sci > 1 or 0 > Sai > 1 or 0 > Sgu > 1 :
        raise ValueError ("All Sij values (as Sug) should be a number from 0 to 1") 

    if bacteria:
        if 0 > Sal > 1:
            raise ValueError ("All Sij values (as Sug) should be a number from 0 to 1") 


    # Sij = Scodon:anticodon
    # Sij of WC interaction is equal to 0
    Sij = { "Sui":0, "Scg":0, "Sau":0, "Sgc":0, 
            "Sug":Sug, "Sci":Sci, "Sai":Sai, 
            "Sgu":Sgu }

    if bacteria: #Lysidine (L) is a bacterial RNA modification of the DNA nucleotide cytidine interact with (A)
        try:
            if "CAU" in dict_anticodon_number["AUA"]:
                Sij["Sal"] = Sal
            else:
                dict_anticodon_number["AUA"].update( { "CAU":0 } )
        except:
            pass

    Wi_dict = {}
    for i_codon in dict_anticodon_number:
        if i_codon[2] == "U":

            #get the anticodon with I in first pos. for the codon with U in first pos.
            key_I = [j for j in dict_anticodon_number[i_codon] if j[0] == "I"]
            key_I = "".join(key_I)

            #get the anticodon with G in first pos. for the codon with U in first pos. (wobble pos)
            key_G = [j for j in dict_anticodon_number[i_codon] if j[0] == "G"]
            key_G = "".join(key_G)

            if key_I != "" and key_G != "":
                Wu = ( ( 1 - Sij["Sui"] ) * dict_anticodon_number[i_codon][key_I] ) +  ( ( 1 - Sij["Sug"] ) *  dict_anticodon_number[i_codon][key_G] ) 
            elif key_I != "" and key_G == "":
                Wu = ( ( 1 - Sij["Sui"] ) * dict_anticodon_number[i_codon][key_I] ) 
            elif key_I == "" and key_G != "":
                Wu = ( ( 1 - Sij["Sug"] ) *  dict_anticodon_number[i_codon][key_G] ) 
            else:
                Wu = 0

            Wi_dict[i_codon] = Wu

        elif i_codon[2] == "C":

            #get the anticodon with G in first pos. for the codon with C in first pos.
            key_G = [j for j in dict_anticodon_number[i_codon] if j[0] == "G"]
            key_G = "".join(key_G)
            #get the anticodon with I in first pos. for the codon with C in first pos. (wobble pos)
            key_I = [j for j in dict_anticodon_number[i_codon] if j[0] == "I"]
            key_I = "".join(key_I)

            if key_I != "": # key_G must be found
                Wc = ( ( 1 - Sij["Scg"] ) * dict_anticodon_number[i_codon][key_G] ) +  ( ( 1 - Sij["Sci"] ) *  dict_anticodon_number[i_codon][key_I] ) 
            else: #if key)I == ""
                Wc = ( ( 1 - Sij["Scg"] ) * dict_anticodon_number[i_codon][key_G] ) 

            Wi_dict[i_codon] = Wc
            
        elif i_codon[2] == "A":

            if i_codon == "AUA":

                if bacteria: 

                    #get the anticodon with U in first pos. for the codon with A in first pos.
                    key_U = [j for j in dict_anticodon_number[i_codon] if j[0] == "U"]
                    key_U1 = "".join(key_U[0])
                    try:
                        key_U2 = "".join(key_U[1])
                    except:
                        key_U2 = False

                    #get the anticodon with I in first pos. for the codon with A in first pos. (wobble pos)
                    key_I = [j for j in dict_anticodon_number[i_codon] if j[0] == "I"]   
                    key_I = "".join(key_I)


                    if key_U2 == True and key_I != "" :
                        Wa = ( ( 1 - Sij["Sau"] ) * dict_anticodon_number[i_codon][key_U1] ) +  ( ( 1 - Sij["Sai"] ) *  dict_anticodon_number[i_codon][key_I] ) + ( ( 1 - Sij["Sal"] ) *  dict_anticodon_number[i_codon][key_U2] )
                    elif key_U2 == False and key_I != "" :
                        Wa = ( ( 1 - Sij["Sau"] ) * dict_anticodon_number[i_codon][key_U1] ) +  ( ( 1 - Sij["Sai"] ) *  dict_anticodon_number[i_codon][key_I] ) 
                    elif key_U2 == True and key_I == "":
                        Wa = ( ( 1 - Sij["Sau"] ) * dict_anticodon_number[i_codon][key_U1] ) + ( ( 1 - Sij["Sal"] ) *  dict_anticodon_number[i_codon][key_U2] )
                    elif key_U2 == False and key_I == "":
                        Wa = ( ( 1 - Sij["Sau"] ) * dict_anticodon_number[i_codon][key_U1] ) 
                
                    Wi_dict[i_codon] = Wa
               

                if not bacteria:
                    #get the anticodon with U in first pos. for the codon with A in first pos.
                    key_U = [j for j in dict_anticodon_number[i_codon] if j[0] == "U"]
                    key_U = "".join(key_U)
                    #get the anticodon with I in first pos. for the codon with A in first pos. (wobble pos)
                    key_I = [j for j in dict_anticodon_number[i_codon] if j[0] == "I"]
                    key_I = "".join(key_I)
                    
                    if key_I != "":
                        Wa = ( ( 1 - Sij["Sau"] ) * dict_anticodon_number[i_codon][key_U] ) +  ( ( 1 - Sij["Sai"] ) *  dict_anticodon_number[i_codon][key_I] ) 
                    else:
                        Wa = ( ( 1 - Sij["Sau"] ) * dict_anticodon_number[i_codon][key_U] ) 

                    Wi_dict[i_codon] = Wa

            else:
                #get the anticodon with U in first pos. for the codon with A in first pos.
                key_U = [j for j in dict_anticodon_number[i_codon] if j[0] == "U"]
                key_U = "".join(key_U)
                #get the anticodon with I in first pos. for the codon with A in first pos. (wobble pos)
                key_I = [j for j in dict_anticodon_number[i_codon] if j[0] == "I"]
                key_I = "".join(key_I)
                
                if key_I != "":
                    Wa = ( ( 1 - Sij["Sau"] ) * dict_anticodon_number[i_codon][key_U] ) +  ( ( 1 - Sij["Sai"] ) *  dict_anticodon_number[i_codon][key_I] ) 
                else:
                    Wa = ( ( 1 - Sij["Sau"] ) * dict_anticodon_number[i_codon][key_U] ) 

                Wi_dict[i_codon] = Wa

            
        elif i_codon[2] == "G":

            #get the anticodon with C in first pos. for the codon with G in first pos.
            key_C = [j for j in dict_anticodon_number[i_codon] if j[0] == "C"]
            key_C = "".join(key_C)
            #get the anticodon with U in first pos. for the codon with G in first pos. (wobble pos)
            key_U = [j for j in dict_anticodon_number[i_codon] if j[0] == "U"]
            key_U = "".join(key_U)

            if key_U != "":
                Wg = ( ( 1 - Sij["Sgc"] ) * dict_anticodon_number[i_codon][key_C] ) +  ( ( 1 - Sij["Sgu"] ) *  dict_anticodon_number[i_codon][key_U] ) 
            else:
                Wg = ( ( 1 - Sij["Sgc"] ) * dict_anticodon_number[i_codon][key_C] ) 

            Wi_dict[i_codon] = Wg

    return Wi_dict

########################################################
              #########################
########################################################

def rel_Wi(dict_abs_Wi,genetic_code_number=1):
    """
    Calculate the relative adaptiveness values for each codon.

    Args:

        dict_abs_Wi (dict): dictionary of the absolute adaptiveness values for each codon returned from abs_Wi() function

        genetic_code_number (int): default = 1, The Genetic Codes number described by NCBI (https://www.ncbi.nlm.nih.gov/Taxonomy/Utils/wprintgc.cgi)

    Returns:
        
        a dictionary of the relative adaptiveness values for each codon

    Raises:

        ValueError if the length of dict_abs_Wi equal to zero

    Example:

        > anticodon_dict = tRNADB_CE("trna.ie.niigata-u.ac.jp/cgi-bin/trnadb/whole_anticodon.cgi?GID=|CP001631&DTYPE=CMP&VTYPE=1")
        # Return an anticodon table of Acidimicrobium ferrooxidans DSM 10331
        > anticodon_codon = dict_codon_anticodon ( anticanticodon_dictodon )
        > dict_codon_anticodon_count = dict_codon_anticodon_count(anticodon_codon,anticodon_dict,bacteria = True)
        > abs_Wi = abs_Wi(dict_codon_anticodon_count, Sug=1, Sci=1, Sai=1, Sgu=1, Sal=1, bacteria=True) 
        > rel_Wi(abs_Wi, 11) 
    """

    if len(dict_abs_Wi) == 0:
        raise ValueError ("dict_abs_Wi can not be with length equal to zero")



    codons = ["".join(i) for i in itertools.product('AUGC', repeat=3)]

    dict_rel_Wi = {}

    # rel_wi if Wi != 0
    max_Wi = max (list(dict_abs_Wi.values()))
    for i_codon in codons:
        if i_codon in dict_abs_Wi and dict_abs_Wi[i_codon] != 0:
            dict_rel_Wi[i_codon] = dict_abs_Wi[i_codon] / max_Wi


    #if stop codons or  remove it
    def get_stop_codons(genetic_code_number):
        from Bio.Data import CodonTable
        genetic_table = CodonTable.unambiguous_dna_by_id[genetic_code_number]
        stop_codons = genetic_table.stop_codons
        stop_M_codons = [i.replace("T","U") for i in stop_codons]
        return stop_M_codons
    

    #remove stop codon from the analysis brefore calc. the geo_mean
    for i_codons_remove in get_stop_codons(genetic_code_number):
        if i_codons_remove in dict_rel_Wi:
            del dict_rel_Wi[i_codons_remove]
        else:
            pass

    #rel_wi if Wi == 0
    def geo_mean(iterable):
    
        #function to calculate the geometric mean
        #iterable = list of number
    
        import numpy as np
        a = np.log(iterable)
        return np.exp(a.sum()/len(a))

    #add the geo_mean if Wi == 0
    #calc geo_mean for rel_wi with Wi != 0
    wi_not_zero = [i for i in list(dict_rel_Wi.values()) if i != 0]
    geo_rel_wi = geo_mean(wi_not_zero)
    for i_codon in codons:
        if i_codon not in dict_rel_Wi or dict_rel_Wi[i_codon] == 0 :
            dict_rel_Wi[i_codon] = geo_rel_wi


    #remove stop codon  from the analysis after calculting the geo_mean
    for i_codons_remove in get_stop_codons (genetic_code_number):
        if i_codons_remove in dict_rel_Wi:
            del dict_rel_Wi[i_codons_remove] 
        else:
            pass

            
    return dict_rel_Wi

########################################################
              #########################
########################################################

def calc_Tai(DNA,rel_dict_wi,genetic_code_number=1):
    """
    Calculate the tRNA adaptation index of a gene.

    Args:

        DNA (str): a coding sequence of DNA ( should only contain A, C, T, and G )

        rel_dict_wi (dict): dictionary of the relative adaptiveness values for each codon returned from rel_Wi() function

        genetic_code_number (int): default = 1, The Genetic Codes number described by NCBI (https://www.ncbi.nlm.nih.gov/Taxonomy/Utils/wprintgc.cgi)

    Returns:
        
        the tRNA adaptation index of a gene

    Raises:

        ValueError if the length of dict_anticodon_number equal to zero

    Example:

        > anticodon_dict = tRNADB_CE("trna.ie.niigata-u.ac.jp/cgi-bin/trnadb/whole_anticodon.cgi?GID=|CP001631&DTYPE=CMP&VTYPE=1")
        # Return an anticodon table of Acidimicrobium ferrooxidans DSM 10331
        > anticodon_codon = dict_codon_anticodon ( anticanticodon_dictodon )
        > dict_codon_anticodon_count = dict_codon_anticodon_count(anticodon_codon,anticodon_dict,bacteria = True)
        > abs_Wi = abs_Wi(dict_codon_anticodon_count, Sug=1, Sci=1, Sai=1, Sgu=1, Sal=1, bacteria=True) 
        > rel_Wi(abs_Wi, 11) 
    """

    if len(rel_dict_wi) == 0:
        raise ValueError ("rel_dict_wi can not be with length equal to zero")

    if DNA.count("A") + DNA.count("T") + DNA.count("C") + DNA.count("G") != len(DNA):
        raise ValueError("The DNA can not include any nucleotides expect (A, T, C, and G)")

    codon_DNA = [ DNA[i:i+3].replace("T","U") for i in range(0,len(DNA),3) if len( DNA[i:i+3] ) == 3  ]
    
    if codon_DNA[0] == "AUG": #start codon remove it
        codon_DNA = codon_DNA[1:]
    else:
        pass

    #if stop codons  remove it
    def get_stop_codons(genetic_code_number):
        from Bio.Data import CodonTable
        genetic_table = CodonTable.unambiguous_dna_by_id[genetic_code_number]
        stop_codons = genetic_table.stop_codons
        stop_M_codons = [i.replace("T","U") for i in stop_codons]
        return stop_M_codons


    stop_codons  = get_stop_codons(genetic_code_number)
    codon_wi = [ rel_dict_wi[i_codon] for i_codon in codon_DNA if i_codon not in stop_codons   ]

    def geo_mean(iterable):
        
        #function to calculate the geometric mean
        #iterable = list of number
        
        import numpy as np
        a = np.log(iterable)
        return np.exp(a.sum()/len(a))
    if sum(codon_wi) != 0:
        Tai_result = geo_mean(codon_wi)
    else:
        Tai_result = 0

    return Tai_result



#Note it will not take genes with N,R,Y,M,k,S,W in the analysis
#################################################################

from gtAI import new_flow
from gtAI import bygaft
from Bio import SeqIO
from gtAI import gtAI
import pandas as pd 
from pandas import DataFrame
#from scipy import stats

def gtai_analysis(main_fasta, GtRNA, genetic_code_number=11, size_pop=60, generation_number=100, bacteria=True):
    """

    Args:

        main_fasta (str): A main fasta file contains genes want to be analyzed (CDS)
        GtRNA (dict): the tRNA genes count
        ref_fasta (str): A reference genes with the highest gene expression in a genome (CDS)
        genetic_code_number (int): default = 1, The Genetic Codes number described by NCBI (https://www.ncbi.nlm.nih.gov/Taxonomy/Utils/wprintgc.cgi)
        size_pop (int): A parameter for the genetic algorithm to identify the population size containing the possible solutions to optimize Sij-values
        generation_number (int): A parameter for the genetic algorithm to identify the generation number
        bacteria (bool): True If the tested organism is prokaryotic or archaea, else equal to False ( default = False )

    Returns:

         df_tai (dataframe): Contains each gene id and its gtAI value 
         final_dict_wi (dict): Contains each codon and its absolute adaptiveness value
         rel_values (dict): Contains each codon and its relative adaptiveness values

    Note: 
        All genes expected to be CDS.
        This function will generate Two files one for ENc values ( a CSV file ) and the other one named best_fit (Contains the optimization information for Sij-values)

    """


    # 1- Calculate the ENc values, and extract the reference set of genes.
    Enc_low_values = new_flow.ENc_calc(main_fasta)

    # 2- Calculate the RSCU values for the reference set.
    RSCU_values = new_flow.RSCU_calc(main_fasta,Enc_low_values,genetic_code_number)
    
    #3- Optimize Sij-values.
    bygaft.gene_algo_corr(dict_tGCN=GtRNA , genetic_code_number = genetic_code_number, RSCU_df=RSCU_values, size_pop=size_pop, generation_number=generation_number, bacteria=bacteria )


    #thie function will return a file named best_fit.py ( must be in the same path )
    while True:
        try:
            import best_fit; best_fit_ = best_fit.best_fit
            steps, variants, fits = list(zip(*best_fit_))
            Sij = variants[-1]
            Sug = Sij[0]
            Sci = Sij[1]
            Sai = Sij[2]
            Sgu = Sij[3]
            Sal = Sij[4]
            if best_fit:
                break
        except:
            continue
    

    # 4- get the tRNA gene copy number ( from GtRNAdb database for example ).
    dict1 = gtAI.dict_codon_anticodon(GtRNA)

    dict2 = gtAI.dict_codon_anticodon_count(dic_codon_anticodon = dict1, dict_tGCN_main = GtRNA, bacteria= bacteria)

    # 5- Calculate the absolute adaptiveness values.
    final_dict_wi = gtAI.abs_Wi(Sug=Sug,Sci=Sci,Sai=Sai,Sgu=Sgu,Sal=Sal,dict_anticodon_number=dict2,bacteria=bacteria)

    # 6- Calculate the relative adaptiveness values.
    rel_values = gtAI.rel_Wi(final_dict_wi,genetic_code_number=genetic_code_number)

    ##########################################################################################################

    
    #7- Calculate the gtAI for all genes.
    tai_data = {}
    for dna in SeqIO.parse(main_fasta, "fasta"):
        dna.seq = dna.seq.upper()
        dna_seq = str(dna.seq)
        if "W" not in dna.seq and "S" not in dna.seq and "N" not in dna.seq and "R" not in dna.seq and "Y" not in dna.seq and "M" not in dna.seq and "K" not in dna.seq:
            TAi = gtAI.calc_Tai(dna_seq, rel_values, genetic_code_number=genetic_code_number)
            tai_data[dna.description] = TAi
        else:
            pass

    df_tai = pd.DataFrame(list(tai_data.items()))
    df_tai.columns = ["id","gtAI"]

    return df_tai , final_dict_wi, rel_values




def RSCU(allseq,allseq_name,The_Genetic_Codes_number):
    """calculate RSCU values

    Args:
        
        allseq (str): DNA sequence
        allseq_name (str) : gene name
        The_Genetic_Codes_number (int) : default = 1, The Genetic Codes number described by NCBI (https://www.ncbi.nlm.nih.gov/Taxonomy/Utils/wprintgc.cgi)

        
    Returns:
        DataFrame: DataFrame contains codons and RSCU values
    """
    from Bio.Data import CodonTable
    from Bio.Seq import Seq
    import re
    from Bio.Alphabet import generic_dna
    import pandas as pd
    from pandas import DataFrame
    from itertools import tee

    xcodontable = CodonTable.unambiguous_dna_by_id[1]
    ycodontable = xcodontable.forward_table
    zcodontable = [ycodontable[i] for i in ycodontable]
    qcodontable = [i for i in ycodontable ]


    for i in zcodontable:
        if zcodontable.count(i) == 1:
            zcodontable.remove(i)
    RSCU = {}

    sequ = str(allseq)

    allseqstr, allseqstr_1  = tee(sequ[i: i+3] for i in range(0, len(sequ), 3) if len(sequ[i: i+3]) == 3)

    qcodontable = ( i for i in qcodontable)
    dic2 = {}
    allseqstr2 = Seq('', generic_dna)
    for i in allseqstr:
        allseqstr2 += i
    aminoacid2 = allseqstr2.translate(table = The_Genetic_Codes_number , stop_symbol ='')
    aminoacid2 = str(aminoacid2)
    RSCUall = {}

    for ii in allseqstr_1:
        dic2[ii] = dic2.get(ii,0) + 1
    for k in  qcodontable:
        RSCUall[k] = 0
        if k in dic2:
            try:
                rscu2 = dic2[k] / ((1/ zcodontable.count(ycodontable[k]))*(aminoacid2.count(ycodontable[k])))
                RSCUall[k] = round(rscu2,6)

            except ZeroDivisionError:
                pass

    df = pd.DataFrame(index=pd.Series([i for i in RSCUall]))
    df[allseq_name] = [RSCUall[r] for r in RSCUall ]
    df.drop(['ATG'],0,inplace= True)
    df.sort_index(inplace=True)
    return df



from gtAI import ENc
from Bio import SeqIO
import pandas as pd 
from pandas import DataFrame
from pathlib import Path 
import math
from gtAI import CA_RSCU  
from gtAI import gtAI

def ENc_calc(fasta_file):
    #######################################################################
    #calculate the Enc values for all genes
    ###############################

    ENc.ENcfilename(fasta_file)

    Enc_file = pd.read_csv(fasta_file + ".enc", delimiter="\t")


    #get the 5% of lwo enc values ( high CUB )
    Enc_low_values = round(math.floor(len(Enc_file) * (5/100)),0)

    if Enc_low_values <= 0:
        Enc_low_values = 1
    else:
        pass
    Enc_low_values = Enc_file.nsmallest(Enc_low_values, 'eq5sun')

    return Enc_low_values
###########################################################################
###########################################################################

def RSCU_calc(fasta_file,Enc_low_values,genetic_code_number):  
    ###########################################################################
    # RSCU values for low and high ENC
    #############################

    #read fasta file
    dna_low_enc = ""
    dna_high_enc = ""
    for dna in SeqIO.parse(fasta_file, "fasta"):

        if str(dna.id) in list(Enc_low_values["id"]):
            dna_low_enc = dna_low_enc + str(dna.seq)



    RSCU_dataframe_low_enc = CA_RSCU.RSCU(dna_low_enc,"low_ENc_values",genetic_code_number)

    #join both dataframe

    return RSCU_dataframe_low_enc
###########################################################################
###########################################################################

def wi_tai_calc(dict_tGCN,Sug, Sci, Sai, Sgu, Sal, genetic_code_number=1,bacteria=False):
    ############################################# ##############################
    # calcularte the Wi wight for the codons and return the corr. between the high/low Enc with the Wi of each codon
    #################################

    #1. get the tRNA number

    #dict_tGCN = tai.get_trna_info(tRNA_url, url_start_from)

    dict1 = gtAI.dict_codon_anticodon(dict_tGCN)

    dict2 = gtAI.dict_codon_anticodon_count(dic_codon_anticodon = dict1, dict_tGCN_main = dict_tGCN, bacteria= bacteria)

    final_dict_wi = gtAI.abs_Wi(Sug=Sug,Sci=Sci,Sai=Sai,Sgu=Sgu,Sal=Sal,dict_anticodon_number=dict2,bacteria=bacteria)

    rel = gtAI.rel_Wi(final_dict_wi,genetic_code_number=genetic_code_number)
    new_rel = {}
    for i_codon in rel: # cuz RSCU value is T not U
        i_codon_new = i_codon.replace("U","T")
        new_rel[i_codon_new] = rel[i_codon]

    #convert the rel to dataframe
    df_rel = pd.DataFrame(list(new_rel.items()))
    df_rel.set_index(df_rel.columns[0],drop=True,inplace=True)
    df_rel.columns = ["Wi"]

    return df_rel
    ###########################################################################
    ###########################################################################


def corr_result(df_rel,RSCU_dataframe_low_enc):
    #merg the Wi.rel and the RSCU values
    RSCU_dataframe_low_enc = RSCU_dataframe_low_enc.merge(df_rel, how='outer', left_index=True, right_index=True)

    RSCU_wai_corr = RSCU_dataframe_low_enc.dropna(axis=0)

    return (RSCU_wai_corr.corr(method="spearman"))

     ###########################################################################
     ###########################################################################

  




def ENcfilename(file):
    import sys
    import argparse
    from Bio import SeqIO
    from pathlib import Path

   
    file_out_put = file + ".enc"



    parser = argparse.ArgumentParser(usage='%(prog)s [options]',
                                     description='Extracts codonusage from CDS input FASTA file. Output will be raw codon counts (.codoncnt), global ACTG counts (.actgcnt), first (.firstcnt), second (.secondcnt), third (.third) codon position counts and Relative Synonymous Codon Usage (.rscucnt). Optional different methods can be applied to calculate Effective Number of Codons (.enc).')
    parser.add_argument("-v", "--verbose", help="increase output verbosity", action="store_true")
    parser.add_argument('-i', help='specify CDS input file in FASTA format' , default = None)
    parser.add_argument('-o', help='specify output prefix',default = None)
    parser.add_argument('-r', action='store_true',
                        help='specify if CDS sequences with length modulo 3 unequal to 0 should be removed and reported to std.out')
    parser.add_argument('-enc', choices=['eq4Wright', 'eq2Sun', 'eq5Sun', 'all'], default = "eq5Sun",
                        help='specify equation to calculate ENC. Either equation (4) [eq4Wright] of (Wright. (1990) Gene 87:23-29) or equation (2) [eq2Sun] or equation (5) [eq5Sun] of (Sun et al. (2012) Mol. Biol. Evol. 30:191-196) or [all].')
    parser.add_argument('-six2fourtwo', default='False', choices=['True', 'False'],
                        help='specify if sixfold codons should be grouped into one fourfold and one twofold group [default: False]. This will only affect calculation of ENC values.')
    args = parser.parse_args()

############
    infile = file
############

#############
    outfile_enc = file_out_put 
############

    six2fourtwo = 'False'
    if args.six2fourtwo == 'True':
        six2fourtwo = True
    if args.six2fourtwo == 'False':
        six2fourtwo = False
    codontable = codontable1
    if six2fourtwo:
        codontable = codontable2

    print("read fasta")
    original_fasta = SeqIO.parse(infile, "fasta")
    print("extract codon counts")
    global_codons = codontable()
    global_actg = actgtable()
    global_first = actgtable()
    global_second = actgtable()
    global_third = actgtable()
    global_rscu = codontable()
    ids_mo3 = []

    if args.enc is not None:
        enchandle = open(outfile_enc, "w")
        if args.enc == 'eq4Wright':
            enchandle.write("id\tlen\tmo3\teq4Wright\n")
        if args.enc == 'eq2Sun':
            enchandle.write("id\tlen\tmo3\teq2Sun\n")
        if args.enc == 'eq5Sun':
            enchandle.write("id\tlen\tmo3\teq5sun\n")
        if args.enc == 'all':
            enchandle.write("id\tlen\tmo3\teq4Wright\teq2Sun\teq5Sun\n")

        c = 0
        cmo3 = 0
        for record in original_fasta:
            c += 1
            tmp_counts = codontable()
            tmp_actgcounts = actgtable()
            tmp_first = actgtable()
            tmp_second = actgtable()
            tmp_third = actgtable()
            tmp_id = record.id
            tmp_len = len(record)
            tmp_mo3 = 0
            if len(record) % 3 != 0:
                tmp_mo3 = 1
                cmo3 += 1
                ids_mo3.append(tmp_id)
                if args.r:
                    continue
            for i in range(0, len(record), 3):
                codon = str(record[i:i + 3].seq)
                for nuc in codon:
                    if nuc in global_actg:
                        global_actg[nuc][1] += 1
                        tmp_actgcounts[nuc][1] += 1
                if len(codon) != 3:
                    continue
                if codon in global_codons:
                    global_codons[codon][2] += 1
                    tmp_counts[codon][2] += 1
                    global_first[codon[0]][1] += 1
                    tmp_first[codon[0]][1] += 1
                    global_second[codon[1]][1] += 1
                    tmp_second[codon[1]][1] += 1
                    global_third[codon[2]][1] += 1
                    tmp_third[codon[2]][1] += 1
                if codon not in global_codons:
                    global_codons['XXX'][2] += 1
                    tmp_counts['XXX'][2] += 1
                    global_first['X'][1] += 1
                    tmp_first['X'][1] += 1
                    global_second['X'][1] += 1
                    tmp_second['X'][1] += 1
                    global_third['X'][1] += 1
                    tmp_third['X'][1] += 1
            if args.enc is not None:
                tmp_enc = {}
                tmp_gcbypos = gcbypos(tmp_counts, six2fourtwo)
                if args.enc == 'eq4Wright':
                    enchandle.write(tmp_id + "\t" + str(tmp_len) + "\t" + str(tmp_mo3) + "\t" + str(
                        calc_eq4wright(tmp_gcbypos[2])) + "\n")
                if args.enc == 'eq2Sun':
                    enchandle.write(tmp_id + "\t" + str(tmp_len) + "\t" + str(tmp_mo3) + "\t" + str(
                        calc_eq2sun(tmp_counts, six2fourtwo)) + "\n")
                if args.enc == 'eq5Sun':
                    enchandle.write(tmp_id + "\t" + str(tmp_len) + "\t" + str(tmp_mo3) + "\t" + str(
                        calc_eq5sun(tmp_counts, six2fourtwo)) + "\n")
                if args.enc == 'all':
                    enchandle.write(tmp_id + "\t" + str(tmp_len) + "\t" + str(tmp_mo3) + "\t" + str(
                        calc_eq4wright(tmp_gcbypos[2])) + "\t" + str(
                        calc_eq2sun(tmp_counts, six2fourtwo)) + "\t" + str(
                        calc_eq5sun(tmp_counts, six2fourtwo)) + "\n")

    print("finished writing")
    if args.r:
        print ('\n'.join(ids_mo3))


def inversetable():
    inversetablecount = {
        'A': {'GCT': 0, 'GCC': 0, 'GCA': 0, 'GCG': 0},
        'R': {'CGT': 0, 'CGC': 0, 'CGA': 0, 'CGG': 0, 'AGA': 0, 'AGG': 0},
        'N': {'AAT': 0, 'AAC': 0},
        'D': {'GAT': 0, 'GAC': 0},
        'C': {'TGT': 0, 'TGC': 0},
        'Q': {'CAA': 0, 'CAG': 0},
        'E': {'GAA': 0, 'GAG': 0},
        'G': {'GGT': 0, 'GGC': 0, 'GGA': 0, 'GGG': 0},
        'H': {'CAT': 0, 'CAC': 0},
        'I': {'ATT': 0, 'ATC': 0, 'ATA': 0},
        'M': {'ATG': 0},
        'L': {'TTA': 0, 'TTG': 0, 'CTT': 0, 'CTC': 0, 'CTA': 0, 'CTG': 0},
        'K': {'AAA': 0, 'AAG': 0},
        'F': {'TTT': 0, 'TTC': 0},
        'P': {'CCT': 0, 'CCC': 0, 'CCA': 0, 'CCG': 0},
        'S': {'TCT': 0, 'TCC': 0, 'TCA': 0, 'TCG': 0, 'AGT': 0, 'AGC': 0},
        'T': {'ACT': 0, 'ACC': 0, 'ACA': 0, 'ACG': 0},
        'W': {'TGG': 0},
        'Y': {'TAT': 0, 'TAC': 0},
        'V': {'GTT': 0, 'GTC': 0, 'GTA': 0, 'GTG': 0},
        '*': {'TAA': 0, 'TGA': 0, 'TAG': 0}
    }
    return inversetablecount


def codontable1():
    codontablecount = {
        'GCT': ['A', 'four', 0],
        'GCC': ['A', 'four', 0],
        'GCA': ['A', 'four', 0],
        'GCG': ['A', 'four', 0],
        'CGT': ['R', 'six', 0],
        'CGC': ['R', 'six', 0],
        'CGG': ['R', 'six', 0],
        'CGA': ['R', 'six', 0],
        'AGA': ['R', 'six', 0],
        'AGG': ['R', 'six', 0],
        'AAT': ['N', 'two', 0],
        'AAC': ['N', 'two', 0],
        'GAT': ['D', 'two', 0],
        'GAC': ['D', 'two', 0],
        'TGT': ['C', 'two', 0],
        'TGC': ['C', 'two', 0],
        'CAA': ['Q', 'two', 0],
        'CAG': ['Q', 'two', 0],
        'GAA': ['E', 'two', 0],
        'GAG': ['E', 'two', 0],
        'GGT': ['G', 'four', 0],
        'GGC': ['G', 'four', 0],
        'GGA': ['G', 'four', 0],
        'GGG': ['G', 'four', 0],
        'CAT': ['H', 'two', 0],
        'CAC': ['H', 'two', 0],
        'ATT': ['I', 'three', 0],
        'ATC': ['I', 'three', 0],
        'ATA': ['I', 'three', 0],
        'ATG': ['M', 'one', 0],
        'TTA': ['L', 'six', 0],
        'TTG': ['L', 'six', 0],
        'CTT': ['L', 'six', 0],
        'CTC': ['L', 'six', 0],
        'CTA': ['L', 'six', 0],
        'CTG': ['L', 'six', 0],
        'AAA': ['K', 'two', 0],
        'AAG': ['K', 'two', 0],
        'TTT': ['F', 'two', 0],
        'TTC': ['F', 'two', 0],
        'CCT': ['P', 'four', 0],
        'CCC': ['P', 'four', 0],
        'CCA': ['P', 'four', 0],
        'CCG': ['P', 'four', 0],
        'TCT': ['S', 'six', 0],
        'TCC': ['S', 'six', 0],
        'TCA': ['S', 'six', 0],
        'TCG': ['S', 'six', 0],
        'AGT': ['S', 'six', 0],
        'AGC': ['S', 'six', 0],
        'ACT': ['T', 'four', 0],
        'ACC': ['T', 'four', 0],
        'ACA': ['T', 'four', 0],
        'ACG': ['T', 'four', 0],
        'TGG': ['W', 'one', 0],
        'TAT': ['Y', 'two', 0],
        'TAC': ['Y', 'two', 0],
        'GTT': ['V', 'four', 0],
        'GTC': ['V', 'four', 0],
        'GTA': ['V', 'four', 0],
        'GTG': ['V', 'four', 0],
        'TAA': ['*', 'three', 0],
        'TGA': ['*', 'three', 0],
        'TAG': ['*', 'three', 0],
        'XXX': ['_missing', 'none', 0]
    }
    return codontablecount


def codontable2():
    codontablecount = {
        'GCT': ['A', 'four', 0],
        'GCC': ['A', 'four', 0],
        'GCA': ['A', 'four', 0],
        'GCG': ['A', 'four', 0],
        'CGT': ['R', 'four', 0],
        'CGC': ['R', 'four', 0],
        'CGG': ['R', 'four', 0],
        'CGA': ['R', 'dour', 0],
        'AGA': ['R_', 'two', 0],
        'AGG': ['R_', 'two', 0],
        'AAT': ['N', 'two', 0],
        'AAC': ['N', 'two', 0],
        'GAT': ['D', 'two', 0],
        'GAC': ['D', 'two', 0],
        'TGT': ['C', 'two', 0],
        'TGC': ['C', 'two', 0],
        'CAA': ['Q', 'two', 0],
        'CAG': ['Q', 'two', 0],
        'GAA': ['E', 'two', 0],
        'GAG': ['E', 'two', 0],
        'GGT': ['G', 'four', 0],
        'GGC': ['G', 'four', 0],
        'GGA': ['G', 'four', 0],
        'GGG': ['G', 'four', 0],
        'CAT': ['H', 'two', 0],
        'CAC': ['H', 'two', 0],
        'ATT': ['I', 'three', 0],
        'ATC': ['I', 'three', 0],
        'ATA': ['I', 'three', 0],
        'ATG': ['M', 'one', 0],
        'TTA': ['L_', 'two', 0],
        'TTG': ['L_', 'two', 0],
        'CTT': ['L', 'four', 0],
        'CTC': ['L', 'four', 0],
        'CTA': ['L', 'four', 0],
        'CTG': ['L', 'four', 0],
        'AAA': ['K', 'two', 0],
        'AAG': ['K', 'two', 0],
        'TTT': ['F', 'two', 0],
        'TTC': ['F', 'two', 0],
        'CCT': ['P', 'four', 0],
        'CCC': ['P', 'four', 0],
        'CCA': ['P', 'four', 0],
        'CCG': ['P', 'four', 0],
        'TCT': ['S', 'four', 0],
        'TCC': ['S', 'four', 0],
        'TCA': ['S', 'four', 0],
        'TCG': ['S', 'four', 0],
        'AGT': ['S_', 'two', 0],
        'AGC': ['S_', 'two', 0],
        'ACT': ['T', 'four', 0],
        'ACC': ['T', 'four', 0],
        'ACA': ['T', 'four', 0],
        'ACG': ['T', 'four', 0],
        'TGG': ['W', 'one', 0],
        'TAT': ['Y', 'two', 0],
        'TAC': ['Y', 'two', 0],
        'GTT': ['V', 'four', 0],
        'GTC': ['V', 'four', 0],
        'GTA': ['V', 'four', 0],
        'GTG': ['V', 'four', 0],
        'TAA': ['*', 'three', 0],
        'TGA': ['*', 'three', 0],
        'TAG': ['*', 'three', 0],
        'XXX': ['_missing', 'none', 0]
    }
    return codontablecount


def actgtable():
    actgtablecount = {
        'A': ['A', 0],
        'C': ['C', 0],
        'T': ['T', 0],
        'G': ['G', 0],
        'X': ['_missing', 0]
    }
    return actgtablecount


# Wright 1990 eqautions
def calc_fa(tmp):
    counts = [x[2] for x in tmp]
    na = sum(counts)
    return [((na * sum([(x / float(na)) ** 2 for x in counts])) - 1) / (na - 1), sum(counts)]


def calc_frc(fa, nrc):
    return 1 / float(nrc) * float(sum(fa))


# Latex function
# \widehat{N}_{c} = 2 + GC_{(3)} + (\frac{29}{GC^{2}_{(3)} + (1 - GC^{2}_{(3)})})
def calc_eq4wright(fgc):
    return 2 + float(fgc) + (29 / ((fgc ** 2) + ((1 - fgc) ** 2)))


# Sun 2012 equations
# FCF according to equation (2) without pseudocounts for eq2Sun
# F_{CF} = \sum_{i=1}^{m} (\frac {n_{i}}{n})^{2}
def calc_fcf(tmp):
    counts = [x[2] for x in tmp]
    na = sum(counts)
    if na == 0:
        return 'NA', 0
    return sum([(x / float(na)) ** 2 for x in counts]), sum(counts)


# FCF according to equation (3) using pseudocounts for eq5Sun
# F_{CF} = \sum_{i=1}^{m} (\frac {n_{i}+1}{n+m})^{2}
def calc_fcf_(tmp):
    counts = [x[2] for x in tmp]
    pseudocounts = [x + 1 for x in counts]
    na = sum(pseudocounts)
    return sum([(x / float(na)) ** 2 for x in pseudocounts]), sum(pseudocounts)


def calc_eq2sun(codoncounts, six2fourtwo):
    six_fcf = []
    six_nrc = []
    four_fcf = []
    four_nrc = []
    three_fcf = []
    three_nrc = []
    two_fcf = []
    two_nrc = []
    one_fcf = []
    one_nrc = []
    four_aa = []
    three_aa = []
    two_aa = []
    one_aa = []
    ncsix = None
    nc = None
    if not six2fourtwo:
        six_aa = ['S', 'R', 'L']
        four_aa = ['A', 'P', 'T', 'G', 'V']
        three_aa = ['I']
        two_aa = ['C', 'E', 'D', 'F', 'H', 'K', 'N', 'Q', 'Y']
        one_aa = ['M', 'W']
        for aa in six_aa:
            tmp_six = [codoncounts[x] for x in codoncounts.keys() if codoncounts[x][0] == aa]
            six = calc_fcf(tmp_six)
            six_fcf.append(six[0])
            six_nrc.append(six[1])
        # setting codons with no counts to average value 1/6
        six_fcf = [float(1) / float(6) if x == 'NA' else x for x in six_fcf]
        ncsix = len(six_fcf) / (sum(six_fcf) / len(six_fcf))
    if six2fourtwo:
        four_aa = ['A', 'P', 'T', 'G', 'V', 'S', 'R', 'L']
        three_aa = ['I']
        two_aa = ['C', 'E', 'D', 'F', 'H', 'K', 'N', 'Q', 'Y', 'S_', 'R_', 'L_']
        one_aa = ['M', 'W']
    for aa in four_aa:
        tmp_four = [codoncounts[x] for x in codoncounts.keys() if codoncounts[x][0] == aa]
        four = calc_fcf(tmp_four)
        four_fcf.append(four[0])
        four_nrc.append(four[1])
    # setting codons with no counts to average value 1/4
    four_fcf = [float(1) / float(4) if x == 'NA' else x for x in four_fcf]
    ncfour = len(four_fcf) / (sum(four_fcf) / len(four_fcf))
    for aa in three_aa:
        tmp_three = [codoncounts[x] for x in codoncounts.keys() if codoncounts[x][0] == aa]
        three = calc_fcf(tmp_three)
        three_fcf.append(three[0])
        three_nrc.append(three[1])
    # setting codons with no counts to average value 1/3
    three_fcf = [float(1) / float(3) if x == 'NA' else x for x in three_fcf]
    ncthree = len(three_fcf) / (sum(three_fcf) / len(three_fcf))
    for aa in two_aa:
        tmp_two = [codoncounts[x] for x in codoncounts.keys() if codoncounts[x][0] == aa]
        two = calc_fcf(tmp_two)
        two_fcf.append(two[0])
        two_nrc.append(two[1])
    # setting codons with no counts to average value 1/2
    two_fcf = [float(1) / float(2) if x == 'NA' else x for x in two_fcf]
    nctwo = len(two_fcf) / (sum(two_fcf) / len(two_fcf))
    for aa in one_aa:
        tmp_one = [codoncounts[x] for x in codoncounts.keys() if codoncounts[x][0] == aa]
        one = calc_fcf(tmp_one)
        one_fcf.append(one[0])
        one_nrc.append(one[1])
    # setting codons with no counts to average value 1/1
    one_fcf = [float(1) / float(1) if x == 'NA' else x for x in one_fcf]
    ncone = len(one_fcf) / (sum(one_fcf) / len(one_fcf))
    if not six2fourtwo:
        nc = ncone + nctwo + ncthree + ncfour + ncsix
    if six2fourtwo:
        nc = ncone + nctwo + ncthree + ncfour
    return nc


# N_{c} = \frac {K_{1} \times \sum_{i}^{K_{1}} n_{i}}{\sum_{i=1}^{K_{i}}(n_{i} \times F_{CF_{i}})}
#  + \frac {K_{2} \times \sum_{i}^{K_{2}} n_{i}}{\sum_{i=1}^{K_{2}}(n_{i} \times F_{CF_{i}})}
#  + \frac {K_{3} \times \sum_{i}^{K_{3}} n_{i}}{\sum_{i=1}^{K_{3}}(n_{i} \times F_{CF_{i}})}
#  + \frac {K_{4} \times \sum_{i}^{K_{4}} n_{i}}{\sum_{i=1}^{K_{4}}(n_{i} \times F_{CF_{i}})}
#  + \frac {K_{6} \times \sum_{i}^{K_{6}} n_{i}}{\sum_{i=1}^{K_{6}}(n_{i} \times F_{CF_{i}})}
def calc_eq5sun(codoncounts, six2fourtwo):
    six_fcf_nrc = []
    four_fcf_nrc = []
    three_fcf_nrc = []
    two_fcf_nrc = []
    one_fcf_nrc = []
    four_aa = []
    three_aa = []
    two_aa = []
    one_aa = []
    ncsix = None
    nc = None
    if not six2fourtwo:
        six_aa = ['S', 'R', 'L']
        four_aa = ['A', 'P', 'T', 'G', 'V']
        three_aa = ['I']
        two_aa = ['C', 'E', 'D', 'F', 'H', 'K', 'N', 'Q', 'Y']
        one_aa = ['M', 'W']
        for aa in six_aa:
            tmp_six = [codoncounts[x] for x in codoncounts.keys() if codoncounts[x][0] == aa]
            six = calc_fcf_(tmp_six)
            six_fcf_nrc.append(six)
        ncsix = len(six_fcf_nrc) / (sum([x[0] * x[1] for x in six_fcf_nrc]) / (sum([x[1] for x in six_fcf_nrc])))
    if six2fourtwo:
        four_aa = ['A', 'P', 'T', 'G', 'V', 'S', 'R', 'L']
        three_aa = ['I']
        two_aa = ['C', 'E', 'D', 'F', 'H', 'K', 'N', 'Q', 'Y', 'S_', 'R_', 'L_']
        one_aa = ['M', 'W']
    for aa in four_aa:
        tmp_four = [codoncounts[x] for x in codoncounts.keys() if codoncounts[x][0] == aa]
        four = calc_fcf_(tmp_four)
        four_fcf_nrc.append(four)
    ncfour = len(four_fcf_nrc) / (sum([x[0] * x[1] for x in four_fcf_nrc]) / (sum([x[1] for x in four_fcf_nrc])))
    for aa in three_aa:
        tmp_three = [codoncounts[x] for x in codoncounts.keys() if codoncounts[x][0] == aa]
        three = calc_fcf_(tmp_three)
        three_fcf_nrc.append(three)
    ncthree = len(three_fcf_nrc) / (sum([x[0] * x[1] for x in three_fcf_nrc]) / (sum([x[1] for x in three_fcf_nrc])))
    for aa in two_aa:
        tmp_two = [codoncounts[x] for x in codoncounts.keys() if codoncounts[x][0] == aa]
        two = calc_fcf_(tmp_two)
        two_fcf_nrc.append(two)
    nctwo = len(two_fcf_nrc) / (sum([x[0] * x[1] for x in two_fcf_nrc]) / (sum([x[1] for x in two_fcf_nrc])))
    for aa in one_aa:
        tmp_one = [codoncounts[x] for x in codoncounts.keys() if codoncounts[x][0] == aa]
        one = calc_fcf_(tmp_one)
        one_fcf_nrc.append(one)
    ncone = len(one_fcf_nrc) / (sum([x[0] * x[1] for x in one_fcf_nrc]) / (sum([x[1] for x in one_fcf_nrc])))
    if not six2fourtwo:
        nc = ncone + nctwo + ncthree + ncfour + ncsix
    if six2fourtwo:
        nc = ncone + nctwo + ncthree + ncfour
    return nc


# Relative Synonymous Codon Usage
# RSCU_{i,j} = \frac{X_{i,j}}{\frac{1}{n_{i}} \times \sum_{j=1}^{n_{i}} X_{i,j}}
def calc_rscu(codoncounts, codontable):
    rscutable = codontable()
    for codon in rscutable.keys():
        aa = rscutable[codon][0]
        tmp_aa = [[x, codoncounts[x][2]] for x in codoncounts.keys() if codoncounts[x][0] == aa]
        counts = [x[1] for x in tmp_aa]
        na = sum(counts)
        if na == 0:
            rscutable[codon][2] = 'NA'
        else:
            tmp_codon_na = [x[1] for x in tmp_aa if x[0] == codon][0]
            tmp_codon_freq = tmp_codon_na / float(na)
            tmp_codon_rscu = tmp_codon_freq * len(counts)
            rscutable[codon][2] = tmp_codon_rscu
    return rscutable


def gcbypos(codoncounts, six2fourtwo):
    aminoacids = ['A', 'C', 'E', 'D', 'G', 'F', 'I', 'H', 'K', 'M', 'L', 'N', 'Q', 'P', 'S', 'R', 'T', 'W', 'V', 'Y']
    if six2fourtwo:
        aminoacids = ['A', 'C', 'E', 'D', 'G', 'F', 'I', 'H', 'K', 'M', 'L', 'L_', 'N', 'Q', 'P', 'S', 'S_', 'R', 'R_',
                      'T', 'W', 'V', 'Y']
    gcone_gc = []
    gcone_sum = []
    gctwo_gc = []
    gctwo_sum = []
    gcthree_gc = []
    gcthree_sum = []
    gcone = None
    gctwo = None
    gcthree = None
    for aa in aminoacids:
        tmp_aa_one = [[x[0], codoncounts[x][2]] for x in codoncounts.keys() if codoncounts[x][0] == aa]
        tmp_aa_two = [[x[1], codoncounts[x][2]] for x in codoncounts.keys() if codoncounts[x][0] == aa]
        tmp_aa_three = [[x[2], codoncounts[x][2]] for x in codoncounts.keys() if codoncounts[x][0] == aa]
        tmp_aa_one_sum = sum([x[1] for x in tmp_aa_one])
        tmp_aa_two_sum = sum([x[1] for x in tmp_aa_two])
        tmp_aa_three_sum = sum([x[1] for x in tmp_aa_three])
        gc_one = sum([x[1] for x in tmp_aa_one if x[0] == 'G' or x[0] == 'C'])
        gcone_sum.append(tmp_aa_one_sum)
        gcone_gc.append(gc_one)
        gc_two = sum([x[1] for x in tmp_aa_two if x[0] == 'G' or x[0] == 'C'])
        gctwo_sum.append(tmp_aa_two_sum)
        gctwo_gc.append(gc_two)
        gc_three = sum([x[1] for x in tmp_aa_two if x[0] == 'G' or x[0] == 'C'])
        gcthree_sum.append(tmp_aa_three_sum)
        gcthree_gc.append(gc_three)
    if float(sum(gcone_gc)) == 0 and float(sum(gcone_sum)) == 0:
        print ('GCone_gc and GCone_sum zero\n')
        gcone = 0
    if float(sum(gcone_gc)) != 0 or float(sum(gcone_sum)) != 0:
        gcone = float(sum(gcone_gc)) / float(sum(gcone_sum))
    if float(sum(gctwo_gc)) == 0 and float(sum(gctwo_sum)) == 0:
        print ('GCtwo_gc and GCtwo_sum zero\n')
        gctwo = 0
    if float(sum(gctwo_gc)) != 0 or float(sum(gctwo_sum)) != 0:
        gctwo = float(sum(gctwo_gc)) / float(sum(gctwo_sum))
    if float(sum(gcthree_gc)) == 0 and float(sum(gcthree_sum)) == 0:
        print ('GCthree_gc and GCthree_sum zero\n')
        gcthree = 0
    if float(sum(gcthree_gc)) != 0 or float(sum(gcthree_sum)) != 0:
        gcthree = float(sum(gcthree_gc)) / float(sum(gcthree_sum))
    return [gcone, gctwo, gcthree]
